// Gray - A simple ray tracing program
// Copyright (C) 2008-2018 Eric Bechet
//
// See the COPYING file for contributions and license information.
// Please report all bugs and problems to <bechet@cadxfem.org>.
//

#ifndef __FRAMEBUFFER_H_
#define __FRAMEBUFFER_H_

#include <ostream>
#include "ray.h"
#include "scene.h"

class MyGlWindow;

class framebuffer
{
public:
  framebuffer ( Gscene * sc_ ) : sc ( sc_ ),empty ( true ) {}
  virtual ~framebuffer() {}
  virtual Color& operator() ( int i,int j ) =0;
  virtual void makeimage ( void );
  virtual size_t getwidth() =0;
  virtual size_t getheight() =0;
  virtual void resize ( size_t w_,size_t h_ ) =0;
  virtual void clear ( void ) =0;
  virtual bool ok()
  {
    return !empty;
  }
  virtual void get_coords ( float i,float j,Ray &r ) =0;
  virtual Gscene * scene ( void )
  {
    return sc;
  }
private:
  Gscene *sc;
protected :
  bool empty;
};

class basic_framebuffer : public framebuffer
{
public:
  basic_framebuffer ( size_t w_,size_t h_,Gscene * sc_ );
  virtual ~basic_framebuffer()
  {
    if (buffer) delete[] buffer;
  }
  virtual Color& operator() ( int i,int j )
  {
    return buffer[i+j*w ];
  }
  virtual float * getbuffer ( void )
  {
    return ( float * ) buffer;
  }
  virtual size_t getwidth()
  {
    return w;
  }
  virtual size_t getheight()
  {
    return h;
  }
  virtual void resize ( size_t w_,size_t h_ );
  virtual void clear ( void );
  virtual void get_coords ( float i,float j,Ray &r ) =0;
  virtual void save_image ( std::ostream &file );
private :
  Color * buffer;
  size_t w;
  size_t h;
  size_t mem;
};


#endif //__FRAMEBUFFER_H_

// kate: indent-mode cstyle; indent-width 2; replace-tabs on; 
