// Gnurbs - A curve and surface library
// Copyright (C) 2008-2018 Eric Bechet
//
// See the LICENSE file for contributions and license information.
// Please report all bugs and problems to <bechet@cadxfem.org>.
//

#include <iostream>
#include "nlagrange.h"
#include "nspline.h"
#include "nbeziercurve.h"
#include "ncoons.h"
#include "nbspline.h"
#include "nbsplinesurface.h"
#include "nutil.h"
#include "gnurbscallbacks.h"
#include "ndisplay.h"
void initcircle(nlagrange &courbe)
{
  for (int i=0;i< courbe.nb_CP();i++)
  {
    double pos= ((double) i) / (courbe.degree());
    double angle=i*n_pi/ (courbe.degree());
    double random=2* (randr()-0.5) /100;
    npoint val;
    random=0;
//  if (i<(degree-1)/2) random=0.05; else random=0;
//  if (i>(degree-1)/2) random=-0.05;
    if (i== (courbe.nb_CP()-1) /2) random=-0.5;
    val[0]=cos(angle) * (1+random);
    val[1]=sin(angle) * (1+random);
    val[2]=0.;
    val[3]=1.;
    courbe.u(i) =pos;
    courbe.set_CP(i,val);
  }
}

void initcircle2(nspline &courbe)
{
  for (int i=0;i< courbe.nb_CP();i++)
  {

    int add=0;
    if (i> (courbe.nb_CP()-1) /2) add=0;
    double pos=i;//1*((double)i+add)/(courbe.nb_CP()-1);
    double angle=i*n_pi/ (courbe.nb_CP()-1);
    double random=2* (randr()-0.5) /100;
    npoint val;
    random=0;
//  if (i<(degree-1)/2) random=0.05; else random=0;
//  if (i>(degree-1)/2) random=-0.05;
    if (i== (courbe.nb_CP()-1) /2) random=-0.5;
    val[0]=cos(angle) * (1+random);
    val[1]=sin(angle) * (1+random);
    val[2]=0;
    val[3]=1.;
    courbe.u(i) =pos;
    courbe.set_CP(i,val);
  }
//  courbe.compute_natural();

}

void initcircle3(nbeziercurve &courbe)
{
  for (int i=0;i< courbe.nb_CP();i++)
  {
    double pos=1* ((double) i) / (courbe.nb_CP()-1);
    double angle=i*n_pi/ (courbe.nb_CP()-1);
    double random=2* (randr()-0.5) /100;
    npoint val;
    random=0;
//  if (i<(degree-1)/2) random=0.05; else random=0;
//  if (i>(degree-1)/2) random=-0.05;
    if (i== (courbe.nb_CP()-1) /2) random=-0.5;
    val[0]=cos(angle) * (1+random);
    val[1]=sin(angle) * (1+random);
    val[2]=0;
    courbe.set_CP(i,val);
  }
}

void initcircle4(nbspline &courbe)    //os[], double val[][2], int degree,int cps)
{
  for (int i=0;i<courbe.nb_CP();++i)
  {
    double angle=i*n_pi/ ((double)(courbe.nb_CP()-1));
    double random=2* (randr()-0.5) /100;
    npoint val;
    random=0;
//  if (i<(degree-1)/2) random=0.05; else random=0;
//  if (i>(degree-1)/2) random=-0.05;
    if (i== (courbe.nb_CP()-1) /2) random=-0.5;
    val[0]=cos(angle) * (1+random);
    val[1]=sin(angle) * (1+random);
    val[2]=0;
    val[3]=1.;
    courbe.set_CP(i,val);
  }
  int ends=courbe.degree() +1;
  int i;
  for (i=1;i<= (courbe.nb_CP()-1+courbe.degree() +2-2*ends);++i)
  {
    courbe.u(i+ends-1) =i;
  }
  for (int j=0;j<ends;++j)
  {
    courbe.u(j) =0.0;
    courbe.u(courbe.nb_CP() +courbe.degree()-j) =i;
  }
}


void initsurface(nbsplinesurface &surf)
{
  for (int i=0;i<surf.nb_CP_u();++i)
  {
    for (int j=0;j<surf.nb_CP_v();++j)
    {
      double R=1.5;
      double x=i*2/ ((double) surf.nb_CP_u()-1)-1;
      double y=j*2/ ((double) surf.nb_CP_v()-1)-1;
      double rr=x*x+y*y;
      double z=sqrt(R*R-rr)-0.9;
      z=z*z;
      /*      if ((i<=cpsu/2) && (j<=cpsv/2)) z=0;
            z=0;
            if ((i==cpsu-1) && (j==cpsv-1)) z=1;
            if ((i==1) && (j==1)) z=1;
            if ((i==cpsu-1) && (j==1)) z=1;
            if ((i==1) && (j==cpsv-1)) z=1;    */
      npoint p(x,y,z,1.0);
      surf.CP(i,j) =p;
    }
  }
  int endsu=surf.degree_u() +1;
  int du=surf.degree_u();
  int cpsu=surf.nb_CP_u()-1;
  for (int i=0;i<endsu;++i)
  {
    surf.u(i) =0.0;
    surf.u(cpsu+du+1-i) = (cpsu+du+2-2*endsu+1);     //1.0;
  }
  for (int i=1;i<= (cpsu+du+2-2*endsu);++i)
  {
    surf.u(i+endsu-1) =i;   //(1.0*i)/(cpsu+du+3-2*endsu);
  }
  int endsv=surf.degree_v() +1;
  int dv=surf.degree_v();
  int cpsv=surf.nb_CP_v()-1;

  for (int i=0;i<endsv;++i)
  {
    surf.v(i) =0.0;
    surf.v(cpsv+dv+1-i) = (cpsv+dv+2-2*endsv+1);     //1.0;
  }
  for (int i=1;i<= (cpsv+dv+2-2*endsv);++i)
  {
    surf.v(i+endsv-1) =i;   //(1.0*i)/(cpsv+dv+3-2*endsv);
  }
//  posv[3]=posv[4]=1.5;//(1.0*i)/(cpsv+dv+3-2*endsv);
//  posu[4]=posu[5]=posu[6]=2;//(1.0*i)/(cpsv+dv+3-2*endsv);
}


void initsurfaceperiodicuv(nbsplinesurface &surf)
{
  int endsu=surf.degree_u() +1;
  int du=surf.degree_u();
  int cpsu=surf.nb_CP_u()-1;

  int endsv=surf.degree_v() +1;
  int dv=surf.degree_v();
  int cpsv=surf.nb_CP_v()-1;
  int cpsv2=cpsv-dv+1;
  for (int j=0;j<=cpsv;++j)
  {
    double anglev=j*2*n_pi/ ((double)(cpsv2));

    int cpsu2=cpsu-du+1;
    double random2=cos(anglev*6) /6;
    for (int i=0;i<=cpsu;++i)
    {
      double angle=i*2*n_pi/ ((double)(cpsu2));
//      double zz=j*2/((double)(cpsv))-1;
      double random=2* (randr() +randr()-0.5) /7;
      random= (1+cos(anglev*4)) *cos(angle*6) /8;
      npoint p;
      p[0]= (cos(angle+anglev/2) * (0.5+random) +1.5) *cos(anglev) * (1+random2) / (2.5);
      p[1]= (cos(angle+anglev/2) * (0.5+random) +1.5) *sin(anglev) * (1+random2) / (2.5);
      p[2]=1+sin(angle+anglev/2) * (0.5+random) / (2.5);
      p[3]=1.0;
      surf.CP(i,j) =p;
    }
  }

  for (int i=0;i<endsu;++i)
  {
    surf.u(i) =-endsu+i+1;
    surf.u(cpsu+du+1-i) = (cpsu+du+2-2*endsu+1) +endsu-i-1;
  }
  for (int i=1;i<= (cpsu+du+2-2*endsu);++i)
  {
    surf.u(i+endsu-1) =i;
  }
  for (int i=0;i<endsv;++i)
  {
    surf.v(i) =-endsv+i+1;
    surf.v(cpsv+dv+1-i) = (cpsv+dv+2-2*endsv+1) +endsv-i-1;
  }
  for (int i=1;i<= (cpsv+dv+2-2*endsv);++i)
  {
    surf.v(i+endsv-1) =i;
  }
}


int main(void)
{
  data_container data;
  ndisplay display;
  display.allow_pick=true;
  int nCP=7;
  nspline courbe2(nCP);

  nbeziercurve courbe3(nCP);
  nbspline courbe4(nCP,3);
  initcircle2(courbe2);
  initcircle3(courbe3);
  initcircle4(courbe4);
//  courbe4.insertknot(1.5,3);
//  courbe4.saturate();
//  courbe4.extend(-0.5);
//  courbe2.Display(data);
  courbe2.compute_natural();
//  courbe2.compute_FD();
//  courbe2.compute_cardinal(0.0);
//  courbe2.Display(data);
//  courbe3.Display(data);
  courbe3.degree_elevation();
  courbe3.degree_elevation();
  courbe3.degree_elevation();
  courbe4.degree_elevation();
  for (int i=0;i<courbe4.nb_knots();++i) std::cout << courbe4.u(i) << " " ;
  std::cout << std::endl;
  int res=0;
//  res=courbe4.deleteknot(8,4,1e-15);
//    for (int i=0;i<courbe4.nb_knots();++i) std::cout << courbe4.u(i) << " " ;
//  std::cout << std::endl;
//  std::cout << res << " deletions." << std::endl;
//  res=courbe4.deleteknot(10,4,1e-15);
//    for (int i=0;i<courbe4.nb_knots();++i) std::cout << courbe4.u(i) << " " ;
//  std::cout << std::endl;
//  std::cout << res << " deletions." << std::endl;
//  res=courbe4.deleteknot(12,4,1e-15);
//  for (int i=0;i<courbe4.nb_knots();++i) std::cout << courbe4.u(i) << " " ;
//  std::cout << std::endl;

//  std::cout << res << " deletions." << std::endl;


  nbeziercurve left(0),right(0);
  npoint p;
  courbe3.cut(0.51,p,left,right);

  npoint a(0,0,0,1.0),b(1,0,0,1.0),c(1,1,0,1.0),d(0,1,0,1.0);
  npoint ab(0.5,0,0.5,1.0),bc(1,0.5,0.25,1.0),cd(0.5,1,-0.5,1.0),da(0,0.5,-0.25,1.0);
  nbeziercurve cab(0),cbc(0);
  cab.add_CP(a);
  cab.add_CP(ab);
  cab.add_CP((ab*-1+b*2+a) /3);
  cab.add_CP(b);
  cbc.add_CP(b);
  cbc.add_CP(bc);
  cbc.add_CP(c);
  nlagrange ccd(0),cda(0);
  ccd.add_CP(0,d);
  ccd.add_CP(0.5,cd);
  ccd.add_CP(1,c);
  cda.add_CP(0,a);
  cda.add_CP(0.5,da);
  cda.add_CP(1,d);
  ncoons surf(cab,ccd,cda,cbc);
  nbsplinesurface bssurface(17,3,8,2);
//  initsurface(bssurface);
  initsurfaceperiodicuv(bssurface);

  gnurbscallbacks CB;

  CB.add_entity(&courbe2);
  CB.add_entity(&courbe3);
  CB.add_entity(&left);
  CB.add_entity(&right);
  CB.add_entity(&courbe4);
  CB.add_entity(&bssurface);
  CB.add_entity(&cda);
  CB.add_entity(&ccd);
  CB.add_entity(&cbc);
  CB.add_entity(&cab);
  CB.add_entity(&surf);

  display.setcallbacks(&CB);
  CB.draw();
  display.init_data(CB.datas);
  display.display();
  return 0;
}
