// nUtil - An utility Library for gnurbs
// Copyright (C) 2008-2018 Eric Bechet
//
// See the LICENSE file for contributions and license information.
// Please report all bugs and problems to <bechet@cadxfem.org>.
//

#include "fltkdisplay.h"
#include <cstring>
#include <iostream>
#include <FL/Fl.H>
#include <FL/fl_draw.H>
#include <FL/gl.h>
#include <FL/Fl_Gl_Window.H>
#include <FL/Fl_Box.H>
#include <FL/Fl_Double_Window.H>
#include <FL/Fl_Menu.H>
#include <FL/Fl_Menu_Bar.H>
#include <FL/Fl_Choice.H>
#include <FL/fl_ask.H>
#ifdef WIN32
#include "windows.h"
#endif //WIN32
#include <limits>
#include "nglwindow.h"


int fltkdisplay::handle(int event)
{
 return 0;
}


int fltkdisplay::getnumpick(void)
{
  int nb=0;
  if (data)
  {
    for (int n=0;n<data->size();++n)
    {
      nb+=(*data)[n].nb_points()+(*data)[n].nb_lines()+(*data)[n].nb_triangles()+(*data)[n].nb_quads();
    }
  }
  return nb;
} // return max number of pickable elements


void fltkdisplay::draw_model(int mode) // called automatically at every refresh
{
  glDepthMask(GL_TRUE);
  _draw_model(false,mode);// display only solid elements
  if (mode!=DRAW_PICK)
  {
    glDepthMask(GL_FALSE);
    _draw_model(true,mode); // display only transparent elements
    glDepthMask(GL_TRUE);
  }
}


void fltkdisplay::_draw_model(bool transparent,int mode) // draw transparent or solid elements, or all if mode==DRAW_PICK
{
  if (data)
  {
    if (mode==DRAW_PICK)
      glInitNames();
    for (int n=0;n<data->size();++n)
    {
      if (drawsurfaces)
      {
        // draw quads
        for (int blk=0;blk<(*data)[n].getnumpropquads();++blk)
        {
          glEnable(GL_LIGHTING);
          const properties &p=(*data)[n].getpropquads(blk);
          if (mode==DRAW_PICK)
          {
            for (int i=0;i<p.nb_uid;++i) glPushName(p.uids[i]);
            glPushName(4);
            glPushName(-1);// will be overwritten
          }
          int last=(blk+1<(*data)[n].getnumpropquads()) ? (*data)[n].getindexquads(blk+1) : (*data)[n].nb_quads();
          if (((p.c.A!=std::numeric_limits<unsigned char>::max()) != (!transparent))||(mode==DRAW_PICK))
          {
            glColor4ub(p.c.R,p.c.G,p.c.B,p.c.A);

            for (int i=(*data)[n].getindexquads(blk);i!=last;++i)
            {
              const quad &q=(*data)[n].get_quad(i);
              npoint3 A,B,normal;
              A=q.pts[1]-q.pts[0];
              B=q.pts[3]-q.pts[0];
              normal.crossprod(A,B);
              normal.normalize();
              if (mode==DRAW_PICK)
                glLoadName(i);
              glBegin(GL_QUADS);
              glNormal3d(normal.x(),normal.y(),normal.z());
              for (int j=0;j<4;++j)
                glVertex3d(q.pts[j].x(),q.pts[j].y(),q.pts[j].z());
              glEnd();
            }
          }
          if (p.edgeon)
          {
            if (((p.edgecolor.A!=std::numeric_limits<unsigned char>::max()) != (!transparent)) || (mode==DRAW_PICK))
            {
              glDisable(GL_LIGHTING);
              glColor4ub(p.edgecolor.R,p.edgecolor.G,p.edgecolor.B,p.edgecolor.A);
              glLineWidth(p.edgethickness);

              for (int i=(*data)[n].getindexquads(blk);i!=last;++i)
              {
                if (mode==DRAW_PICK)
                  glLoadName(i);
                glBegin(GL_LINES);
                const quad &q=(*data)[n].get_quad(i);
                for (int j=0;j<4;++j)
                {
                  glVertex3d(q.pts[j].x(),q.pts[j].y(),q.pts[j].z());
                  glVertex3d(q.pts[(j+1)%4].x(),q.pts[(j+1)%4].y(),q.pts[(j+1)%4].z());
                }
                glEnd();
              }
            }
          }
          if (mode==DRAW_PICK)
          {
            glPopName();
            glPopName();
            for (int i=0;i<p.nb_uid;++i) glPopName();
          }
        }
      }

      
      // draw triangles
      if (drawsurfaces)
      {
        for (int blk=0;blk<(*data)[n].getnumproptriangles();++blk)
        {
          glEnable(GL_LIGHTING);
          const properties &p=(*data)[n].getproptriangles(blk);
          if (mode==DRAW_PICK)
          {
            for (int i=0;i<p.nb_uid;++i) glPushName(p.uids[i]);
            glPushName(3);
            glPushName(-1);// will be overwritten
          }
          int last=(blk+1<(*data)[n].getnumproptriangles()) ? (*data)[n].getindextriangles(blk+1) : (*data)[n].nb_triangles();
          if (((p.c.A!=std::numeric_limits<unsigned char>::max()) != (!transparent))||(mode==DRAW_PICK))
          {
            glColor4ub(p.c.R,p.c.G,p.c.B,p.c.A);
            for (int i=(*data)[n].getindextriangles(blk);i!=last;++i)
            {
              const triangle &t=(*data)[n].get_triangle(i);
              npoint3 A,B,normal;
              A=t.pts[1]-t.pts[0];
              B=t.pts[2]-t.pts[0];
              normal.crossprod(A,B);
              normal.normalize();
              if (mode==DRAW_PICK)
                glLoadName(i);
              glBegin(GL_TRIANGLES);
              glNormal3d(normal.x(),normal.y(),normal.z());
              for (int j=0;j<3;++j)
                glVertex3d(t.pts[j].x(),t.pts[j].y(),t.pts[j].z());
              glEnd();
            }
          }
          if (p.edgeon&&drawlines)
          {
            glDisable(GL_LIGHTING);
            if (((p.edgecolor.A!=std::numeric_limits<unsigned char>::max()) != (!transparent))||(mode==DRAW_PICK))
            {
              glColor4ub(p.edgecolor.R,p.edgecolor.G,p.edgecolor.B,p.edgecolor.A);
              glLineWidth(p.edgethickness);

              for (int i=(*data)[n].getindextriangles(blk);i!=last;++i)
              {
                if (mode==DRAW_PICK)
                  glLoadName(i);
                glBegin(GL_LINES);
                const triangle &t=(*data)[n].get_triangle(i);
                for (int j=0;j<3;++j)
                {
                  glVertex3d(t.pts[j].x(),t.pts[j].y(),t.pts[j].z());
                  glVertex3d(t.pts[(j+1)%3].x(),t.pts[(j+1)%3].y(),t.pts[(j+1)%3].z());
                }
                glEnd();
              }
            }
          }
          if (mode==DRAW_PICK)
          {
            glPopName();
            glPopName();
            for (int i=0;i<p.nb_uid;++i) glPopName();
          }
        }
      }
      
      // draw lines
      glDisable(GL_LIGHTING);
      if (drawlines)
      {
        for (int blk=0;blk<(*data)[n].getnumproplines();++blk)
        {
          const properties &p=(*data)[n].getproplines(blk);
          if (mode==DRAW_PICK)
          {
            for (int i=0;i<p.nb_uid;++i) glPushName(p.uids[i]);
            glPushName(2);
            glPushName(-1);// will be overwritten
          }
          int last=(blk+1<(*data)[n].getnumproplines()) ? (*data)[n].getindexlines(blk+1) : (*data)[n].nb_lines();
          if (((p.c.A!=std::numeric_limits<unsigned char>::max()) != (!transparent))||(mode==DRAW_PICK))
          {
            glColor4ub(p.c.R,p.c.G,p.c.B,p.c.A);
            glLineWidth(p.thickness);
            for (int i=(*data)[n].getindexlines(blk);i!=last;++i)
            {
              if (mode==DRAW_PICK)
                glLoadName(i);
              glBegin(GL_LINES);
              glNormal3f(0.,0.,0.);
              const line &l=(*data)[n].get_line(i);
              for (int j=0;j<2;++j)
                glVertex3d(l.pts[j].x(),l.pts[j].y(),l.pts[j].z());
              glEnd();
            }
          }
          if (mode==DRAW_PICK)
          {
            glPopName();
            glPopName();
            for (int i=0;i<p.nb_uid;++i) glPopName();
          }
        }
      }
      
      // draw points
      glDisable(GL_LIGHTING);
      if (drawpoints)
      {
        for (int blk=0;blk<(*data)[n].getnumproppoints();++blk)
        {
          const properties &p=(*data)[n].getproppoints(blk);
          if (mode==DRAW_PICK)
          {
            for (int i=0;i<p.nb_uid;++i) glPushName(p.uids[i]);
            glPushName(1);
            glPushName(-1);// will be overwritten
          }
          int last=(blk+1<(*data)[n].getnumproppoints()) ? (*data)[n].getindexpoints(blk+1) : (*data)[n].nb_points();
          if (((p.c.A!=std::numeric_limits<unsigned char>::max()) != (!transparent))||(mode==DRAW_PICK))
          {
            glColor4ub(p.c.R,p.c.G,p.c.B,p.c.A);
            glPointSize(p.pointsize);
            for (int i=(*data)[n].getindexpoints(blk);i!=last;++i)
            {
              if (mode==DRAW_PICK)
                glLoadName(i);
              glBegin(GL_POINTS);
              glNormal3f(0.,0.,0.);
              const point &p=(*data)[n].get_point(i);
              glVertex3d(p.pts.x(),p.pts.y(),p.pts.z());
              glEnd();
            }
          }
          if (mode==DRAW_PICK)
          {
            glPopName();
            glPopName();
            for (int i=0;i<p.nb_uid;++i) glPopName();
          }
        }
      }
      
      //draw texts
      if ((!transparent)||(mode==DRAW_PICK))
      {
        glDisable(GL_LIGHTING);
        
        if (drawtexts)
        {
          gl_font(0, (fontsize*12)/14);
          glColor3f(1.0, 1.0, 1.0);
          for (int dim=0;dim<3;dim++)
          {
            int nb=(*data)[n].nb_texts(dim);
            for (int i=0;i<nb;i++)
            {
              const point &pt=(*data)[n].get_text(dim,i);
              const npoint3 &pos=pt.pts;        // 3D geometric position
              const char *txt=pt.info.c_str();  // actual text string
              glRasterPos3f(pos[0],pos[1],pos[2]);
              gl_draw(txt, strlen(txt));
            }
          }
        }
      }
    }
    glEnable(GL_LIGHTING);
  }
}



fltkdisplay::fltkdisplay(color _c,char *_name) : llf(std::numeric_limits< double >::max(),std::numeric_limits< double >::max(),std::numeric_limits< double >::max()),
                                                urb(std::numeric_limits< double >::min(),std::numeric_limits< double >::min(),std::numeric_limits< double >::min()),
                                                data(NULL),drawlines(true),drawpoints(true),drawsurfaces(true),drawtexts(true),own(false)
{
  if (_name)
  {
    name=new char[strlen(_name)+1];
    strcpy(name,_name);
  }
  else
    name=NULL;
  c=_c;
  allow_pick=false;
  setcallbacks(NULL);
}

void fltkdisplay::reset_bb(void)
{
  llf=npoint3(std::numeric_limits< double >::max(),std::numeric_limits< double >::max(),std::numeric_limits< double >::max());
  urb=npoint3(std::numeric_limits< double >::min(),std::numeric_limits< double >::min(),std::numeric_limits< double >::min());
}



void chk_bb(npoint3 &llf,npoint3 &urb,const npoint3 &t)
{
  for (int i=0;i<3 ; ++i)
  {
    if (llf[i]>t[i]) llf[i]=t[i];
    if (urb[i]<t[i]) urb[i]=t[i];
  }
}

void fltkdisplay::compute_bb(void)
{
  if (data)
  {
    for (int n=0;n<data->size();++n)
    {
      for (int i=0;i!=(*data)[n].nb_quads();++i)
      {
        const quad &q=(*data)[n].get_quad(i);
        for (int j=0;j<4;++j)
        {
          chk_bb(llf,urb,q.pts[j]);
        }
      }
      for (int i=0;i!=(*data)[n].nb_triangles();++i)
      {
        const triangle &t=(*data)[n].get_triangle(i);
        for (int j=0;j<3;++j)
        {
          chk_bb(llf,urb,t.pts[j]);
        }
      }
      for (int i=0;i!=(*data)[n].nb_lines();++i)
      {
        const line &l=(*data)[n].get_line(i);
        for (int j=0;j<2;++j)
        {
          chk_bb(llf,urb,l.pts[j]);
        }
      }
      for (int i=0;i!=(*data)[n].nb_points();++i)
      {
        const point &p=(*data)[n].get_point(i);
        chk_bb(llf,urb,p.pts);
      }
      for (int dim=0;dim<3;dim++)
      {
        int nb=(*data)[n].nb_texts(dim);
        for (int i=0;i<nb;i++)
        {
          const point &pt=(*data)[n].get_text(dim,i);
          chk_bb(llf,urb,pt.pts);
        }
      }
    }
  }
  double sz[3];
  double mean[3];
  int max=-1;
  double ratio=0.001;
  double valmax=-1;
  for (int i=0;i<3;++i)
  {
    sz[i]=fabs(llf[i]-urb[i]);
    mean[i]=(llf[i]+urb[i])/2.0;
    if (valmax<sz[i])
    {
      valmax=sz[i];
      max=i;
    }
  }
  if (valmax==0.0) { valmax=1.0; max=-1;}
  for (int i=0;i<3;++i)
  {
    if (i!=max)
      if (sz[i]<ratio*valmax)
      {
        llf[i]=mean[i]-valmax*ratio/2.0;
        urb[i]=mean[i]+valmax*ratio/2.0;
            sz[i]=fabs(llf[i]-urb[i]);
            mean[i]=(llf[i]+urb[i])/2.0;

      }
  }
}

fltkdisplay::~fltkdisplay()
{
  if (name) delete[] name;
  if (own&&data) delete data;
}


class MyWindow : public Fl_Double_Window {
  public:
    MyWindow(int w, int h, const char* t,fltkdisplay *f) : Fl_Double_Window(w,h,t),fd(f) {}
    ~MyWindow(){}
    fltkdisplay *fd;
};


void quit_callback(Fl_Widget* w, void* p)
{
  if (fl_choice("Are you sure you want to quit?", fl_no, fl_yes, 0))
    exit(0);
}

void drawP_callback(Fl_Widget* w, void* p)
{
    MyWindow *win = (MyWindow *)p;
    fltkdisplay *fd=win->fd;
    fd->drawpoints=!fd->drawpoints;
    fd->gl->redraw();
}

void drawL_callback(Fl_Widget* w, void* p)
{
    MyWindow *win = (MyWindow *)p;
    fltkdisplay *fd=win->fd;
    fd->drawlines=!fd->drawlines;
    fd->gl->redraw();
}

void drawS_callback(Fl_Widget* w, void* p)
{
    MyWindow *win = (MyWindow *)p;
    fltkdisplay *fd=win->fd;
    fd->drawsurfaces=!fd->drawsurfaces;
    fd->gl->redraw();
}

void drawT_callback(Fl_Widget* w, void* p)
{
    MyWindow *win = (MyWindow *)p;
    fltkdisplay *fd=win->fd;
    fd->drawtexts=!fd->drawtexts;
    fd->gl->redraw();
}


Fl_Menu_Item menuitems[] = {
  { "&Draw",              0, 0, 0, FL_SUBMENU },
    { "Toggle &Points",   FL_COMMAND + 'p', (Fl_Callback *)drawP_callback },
    { "Toggle &Lines",    FL_COMMAND + 'l', (Fl_Callback *)drawL_callback},
    { "Toggle &Surfaces",  FL_COMMAND + 's', (Fl_Callback *)drawS_callback},
    { "Toggle &Texts",  FL_COMMAND + 't', (Fl_Callback *)drawT_callback, 0, FL_MENU_DIVIDER },
    { "&Quit", FL_COMMAND + 'q', (Fl_Callback *)quit_callback, 0 },
    { 0 },
 /* { "&Edit", 0, 0, 0, FL_SUBMENU },
    { "&Undo",       FL_COMMAND + 'z', (Fl_Callback *)NULL, 0, FL_MENU_DIVIDER },
    { "Cu&t",        FL_COMMAND + 'x', (Fl_Callback *)NULL },
    { "&Copy",       FL_COMMAND + 'c', (Fl_Callback *)NULL },
    { "&Paste",      FL_COMMAND + 'v', (Fl_Callback *)NULL },
    { "&Delete",     0, (Fl_Callback *)NULL },
    { 0 },*/
  { 0 }
};


int eval_fontsize(int scrW)
{
      if(scrW <= 800) return 10;
      if(scrW <= 1024) return 11;
      if(scrW <= 1440) return 12;
      if(scrW <= 1680) return 13;
      if(scrW <= 1920) return 14;
      if(scrW <= 2304) return 15;
      if(scrW <= 2560) return 16;
      if(scrW <= 2880) return 18;
      if(scrW <= 3200) return 24;
      return ((scrW-800)/160 + 8); 
}

void fltkdisplay::display(bool perspectiveon)
{
  if (ncb)
  {
    const std::vector<data_container> *datas=ncb->get_data();
    if (datas)
    {
      ncb->draw();
      init_data(*datas);
    }
    else
    {
      ncb->draw();
    }
  }
  int scrW=Fl::w();
  int scrH=Fl::h();
  int W=(800*scrW)/1920;
  int H=(600*scrW)/1920;
  int fsize=eval_fontsize(scrW);
  FL_NORMAL_SIZE=fsize;
  fontsize=fsize;
  
#if 0
  int k = Fl::set_fonts(0 ? (0>1 ? "*" : 0) : "-*");
  for (int i = 0; i < k; i++) {
    int t; const char *name = Fl::get_font_name((Fl_Font)i,&t);
    char buffer[128];
    if (t) {
      char *p = buffer;
      if (t & FL_BOLD) {*p++ = '@'; *p++ = 'b';}
      if (t & FL_ITALIC) {*p++ = '@'; *p++ = 'i';}
          *p++ = '@'; *p++ = '.'; // Suppress subsequent formatting - some MS fonts have '@' in their name
      strcpy(p,name);
      name = buffer;
    }
    
    int *s; int n = Fl::get_font_sizes((Fl_Font)i, s);
    std::cout << name ;
    for (int i=0;i<n;++i) std::cout << " "<< s[i];
    std::cout << std::endl;
  }
#endif
  
  
//  Fl_Tooltip::size(FL_NORMAL_SIZE);
  
  MyWindow *win =new MyWindow ( W, H,name,this);
  win->begin();
  Fl_Menu_Bar *m = new Fl_Menu_Bar(0, 0, 0, 0);
  int tsz= m->textsize();
  m->resize(0, 0, win->w(), 2*tsz);
  m->copy(menuitems,win);
  gl = new nglwindow ( 0,m->h(),win->w(), win->h()-2*m->h(),c,this);
/*  Fl_Box *box_t = new Fl_Box(0,0,800,100,"3d view");
  Fl_Box *box = new Fl_Box(0,700,800,100,"Hello, World!");
  Fl_Box *box_s1 = new Fl_Box(0,0,100,800,"side1");
  Fl_Box *box_s2 = new Fl_Box(700,0,100,800,"side2");
  box->box(FL_UP_BOX);
  box->labelfont(FL_BOLD+FL_ITALIC);
  box->labelsize(36);
  box->labeltype(FL_SHADOW_LABEL);*/
  
  

  /*  FL_RGB - RGB color (not indexed)
    FL_RGB8 - RGB color with at least 8 bits of each color
    FL_INDEX - Indexed mode
    FL_SINGLE - not double buffered
    FL_DOUBLE - double buffered
    FL_ACCUM - accumulation buffer
    FL_ALPHA - alpha channel in color
    FL_DEPTH - depth buffer
    FL_STENCIL - stencil buffer
    FL_MULTISAMPLE */
/*
  std::cout << (gl->can_do(FL_RGB)?"FL_RGB ":"") ;
  std::cout << (gl->can_do(FL_RGB8)?"FL_RGB8 ":"");
  std::cout << (gl->can_do(FL_INDEX)?"FL_INDEX ":"");
  std::cout << (gl->can_do(FL_SINGLE)?"FL_SINGLE ":"");
  std::cout << (gl->can_do(FL_DOUBLE)?"FL_DOUBLE ":"");
  std::cout << (gl->can_do(FL_ACCUM)?"FL_ACCUM ":"");
  std::cout << (gl->can_do(FL_ALPHA)?"FL_ALPHA ":"");
  std::cout << (gl->can_do(FL_DEPTH)?"FL_DEPTH ":"");
  std::cout << (gl->can_do(FL_STENCIL)?"FL_STENCIL ":"");
  std::cout << (gl->can_do(FL_MULTISAMPLE)?"FL_MULTISAMPLE ":"") << std::endl;*/
  gl->mode(FL_ALPHA|FL_DOUBLE|FL_ACCUM|FL_DEPTH|FL_MULTISAMPLE);
  double focal=2.8;
  double lens=2.326661531;
  double fov =  (lens/2/focal ) * 2 * 180 / 3.14159265;
  npoint3 center=(llf+urb)/2.0;
  npoint3 size=urb-llf;
  double sz=size.norm();
  double nearclip=sz/10;
  double farclip=sz*10;
//  double size2=sqrt(size[0]*size[0]+size[1]*size[1]);
  const double look[3]={center[0],center[1],center[2]};
  const double vp[3]={center[0],center[1],center[2]+sz*1.5};
  const double up[3]={0.,1.,0.};
  gl->set_view(lens,focal,fov,nearclip,farclip,vp,look,up);
  gl->flush();
  win->end();
  win->resizable ( gl );
  win->show();
  Fl::run();
  delete gl;
  delete win;
}

void fltkdisplay::init_data(const data_container &data_)
{
  std::vector<data_container>*data__=new std::vector<data_container>;
  data__->push_back(data_);
  data=data__;
  own=true;
  reset_bb();
  compute_bb();
}

void fltkdisplay::init_data(const std::vector< data_container >& datas_)
{
  data=&datas_;
  reset_bb();
  compute_bb();
}
