// nUtil - An utility Library for gnurbs
// Copyright (C) 2008-2018 Eric Bechet
//
// See the LICENSE file for contributions and license information.
// Please report all bugs and problems to <bechet@cadxfem.org>.
//

#ifndef __FLTKDISPLAY_H
#define __FLTKDISPLAY_H

#include <vector>

#include "nutil.h"
#include "ndisplay_base.h"

class ngldraw
{
public:
  ngldraw() {}
  virtual void draw_model(int mode=DRAW_NORMAL)=0; // called by the opengl window at every refresh
  virtual int handle(int event) =0; // called by the opengl window at every event
  virtual ~ngldraw(){};
};

class nglwindow;

class fltkdisplay: public ngldraw ,public ndisplay_base
{
protected :

public:
  fltkdisplay(color _c=color(25, 51, 102,255),char *_name=NULL); // constructor
  virtual ~fltkdisplay(); // destructor
  virtual void init_data(const std::vector<data_container> &data);   // all
  virtual void init_data(const data_container &data);   // one dataset - will be copied
  virtual void display(bool perspectiveon=0);   // display loop
  virtual void draw_model( int mode=DRAW_NORMAL); // called by the opengl window at every refresh
  virtual int handle(int event); // called by the opengl window at every event
  virtual int getnumpick(void); // return max number of pickable elements
  bool drawlines;
  bool drawpoints;
  bool drawsurfaces;
  bool drawtexts;
  int fontsize;
  nglwindow *gl;
private :
  virtual void _draw_model(bool transparent, int mode); // draw either solid faces or transparent ones w/o depth update
  virtual void compute_bb();
  virtual void reset_bb();
  const std::vector<data_container> *data;
  color c;
  char *name;
  npoint3 llf,urb;//bounding box
  bool own; //data
};


#endif// __FLTKDISPLAY_H
